require 'sinatra'
require 'sinatra/sse'
require 'awesome_print'
require 'to_slug'
require 'json'
require 'emittr'

require 'ezpaas/helpers/container_manager'
require 'ezpaas/models/app'

module EzPaaS
  module Server
    module Routes
      class Deployments < Sinatra::Application

        include Sinatra::SSE

        post '/' do
          ensure_app
          manager = Helpers::ContainerManager.new
          body = request.body # String IO object containing the source tarball

          open_message_stream do |emitter|
            slug = manager.create_slug(body, emitter)

            emitter.emit :message, '-----> Saving new slug name to database'
            @app.slug = slug

            procfile = manager.read_procfile(@app.slug)
            @app.scale ||= {}
            procfile.keys.each do |key|
              @app.scale[key] ||= 1
            end

            redeploy(emitter)

            @app.save
          end
        end

        patch '/' do
          ensure_app

          scale = request.params['scale']
          if scale.nil? || !scale.is_a?(Hash)
            content_type :json
            halt [400, { error: 'scale parameter missing or malformed' }.to_json]
          end

          @app.scale.keys.each do |key|
            if (newcount = scale[key]) && scale[key] == scale[key].to_i.to_s
              @app.scale[key] = newcount.to_i
            end
          end

          open_message_stream do |emitter|
            redeploy(emitter)
            @app.save
          end
        end

        delete '/' do
          ensure_app

          @app.scale.keys.each do |key|
            @app.scale[key] = 0
          end

          manager = Helpers::ContainerManager.new

          open_message_stream do |emitter|
            scale_strings = @app.scale.map { |k, v| "#{k}: #{v}" }
            emitter.emit :message, "-----> App scaled to: #{scale_strings.join(', ')}"
            manager.undeploy_app(@app.name, emitter)
            @app.save
          end
        end

        private

        def ensure_app
          app_name = request.params['app']

          if app_name.nil?
            app_not_found
          end

          app_slug = app_name.to_slug

          app = Models::App.where(name: app_slug).first

          if app.nil?
            app_not_found
          end

          @app = app
        end

        def open_message_stream
          sse_stream do |out|
            begin
              emitter = Emittr::Emitter.new

              emitter.on :message do |message|
                out.push :event => 'message', :data => message
              end

              yield emitter
            rescue Exception => ex
              raise
            ensure
              out.close
            end
          end
        end

        def redeploy(emitter)
          manager = Helpers::ContainerManager.new

          scale_strings = @app.scale.map { |k, v| "#{k}: #{v}" }
          emitter.emit :message, "-----> App scaled to: #{scale_strings.join(', ')}"

          emitter.emit :message, '-----> Deploying application'
          manager.undeploy_app(@app.name, emitter)
          manager.deploy_app(@app.name, @app.slug, @app.scale, emitter)

          emitter.emit :message, '-----> Application deployed'
        end

        def app_not_found
          content_type :json
          halt [404, { error: 'app not found' }.to_json]
        end

      end

    end
  end
end
